import logging
from collections import defaultdict
from datetime import datetime
from typing import List, Dict

from peek_plugin_diagram._private.tuples.grid.EncodedGridTuple import (
    EncodedGridTuple,
)
from twisted.internet.defer import DeferredList, Deferred, inlineCallbacks
from vortex.DeferUtil import vortexLogFailure
from vortex.Payload import Payload
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexABC import SendVortexMsgResponseCallable
from vortex.VortexFactory import VortexFactory

from peek_abstract_chunked_index.private.client.handlers.ACICacheHandlerABC import (
    ACICacheHandlerABC,
)
from peek_abstract_chunked_index.private.tuples.ACIUpdateDateTupleABC import (
    ACIUpdateDateTupleABC,
)
from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.client.controller.GridCacheController import (
    GridCacheController,
)
from peek_plugin_diagram._private.server.client_handlers.ClientGridLoaderRpc import (
    ClientGridLoaderRpc,
)
from peek_plugin_diagram._private.storage.GridKeyIndex import (
    GridKeyIndexCompiled,
)

logger = logging.getLogger(__name__)

clientGridWatchUpdateFromDeviceFilt = {"key": "clientGridWatchUpdateFromDevice"}
clientGridWatchUpdateFromDeviceFilt.update(diagramFilt)

#: This the type of the data that we get when the clients observe new grids.
DeviceGridT = Dict[str, datetime]


# ModelSet HANDLER
class GridCacheHandler(ACICacheHandlerABC):
    _UpdateDateTuple: ACIUpdateDateTupleABC = GridKeyIndexCompiled
    _updateFromServerFilt: Dict = clientGridWatchUpdateFromDeviceFilt
    _logger: logging.Logger = logger

    def __init__(self, cacheController: GridCacheController, clientId: str):
        """App Grid Handler

        This class handles the custom needs of the desktop/mobile
         apps observing grids.

        """
        ACICacheHandlerABC.__init__(self, cacheController, clientId)

        # We need to know who is watching what so we can tell the server.
        self._observedGridKeysByVortexUuid = defaultdict(list)
        self._observedVortexUuidsByGridKey = defaultdict(list)

        # We're not using this
        del self._uuidsObserving

    # ---------------
    # Filter out offline vortexes

    def _filterOutOfflineVortexes(self):
        # TODO, Change this to observe offline vortexes
        # This depends on the VortexFactory offline observable implementation.
        # Which is incomplete at this point :-|

        vortexUuids = set(VortexFactory.getRemoteVortexUuids())
        vortexUuidsToRemove = (
            set(self._observedGridKeysByVortexUuid) - vortexUuids
        )

        if not vortexUuidsToRemove:
            return

        for vortexUuid in vortexUuidsToRemove:
            del self._observedGridKeysByVortexUuid[vortexUuid]

        self._rebuildStructs()

    # ---------------
    # Process update from the server

    def notifyOfUpdate(self, gridKeys: List[str]):
        """Notify of Grid Updates

        This method is called by the client.GridCacheController when it receives updates
        from the server.

        """
        self._filterOutOfflineVortexes()

        def cratePayloadEnvelope():
            payloadEnvelope = PayloadEnvelope()
            payloadEnvelope.data = []
            return payloadEnvelope

        payloadsByVortexUuid = defaultdict(cratePayloadEnvelope)

        for gridKey in gridKeys:

            gridTuple = self._cacheController.encodedChunk(gridKey)
            if not gridTuple:
                gridTuple = EncodedGridTuple()
                gridTuple.gridKey = gridKey

            vortexUuids = self._observedVortexUuidsByGridKey.get(gridKey, [])

            # Queue up the required client notifications
            for vortexUuid in vortexUuids:
                logger.debug(
                    "Sending unsolicited grid %s to vortex %s",
                    gridKey,
                    vortexUuid,
                )
                payloadsByVortexUuid[vortexUuid].data.append(gridTuple)

        # Send the updates to the clients
        dl = []
        for vortexUuid, payloadEnvelope in list(payloadsByVortexUuid.items()):
            payloadEnvelope.filt = clientGridWatchUpdateFromDeviceFilt

            # Serliase in thread, and then send.
            d = payloadEnvelope.toVortexMsgDefer(base64Encode=False)
            d.addCallback(
                VortexFactory.sendVortexMsg, destVortexUuid=vortexUuid
            )
            dl.append(d)

        # Log the errors, otherwise we don't care about them
        dl = DeferredList(dl, fireOnOneErrback=True)
        dl.addErrback(vortexLogFailure, logger, consumeError=True)

    # ---------------
    # Process observes from the devices

    @inlineCallbacks
    def _processObserve(
        self,
        payloadEnvelope: PayloadEnvelope,
        vortexUuid: str,
        sendResponse: SendVortexMsgResponseCallable,
        **kwargs
    ):
        cacheAll = payloadEnvelope.filt.get("cacheAll") == True

        payload = yield payloadEnvelope.decodePayloadDefer()

        lastUpdateByGridKey: DeviceGridT = payload.tuples[0]

        if not cacheAll:
            gridKeys = list(lastUpdateByGridKey.keys())
            self._observedGridKeysByVortexUuid[vortexUuid] = gridKeys
            self._rebuildStructs()

        self._replyToObserve(
            payload.filt, lastUpdateByGridKey, sendResponse, cacheAll=cacheAll
        )

    def _rebuildStructs(self) -> None:
        """Rebuild Structs

        Rebuild the reverse index of uuids by grid key.

        :returns: None
        """
        # Rebuild the other reverse lookup
        newDict = defaultdict(list)

        for vortexUuid, gridKeys in self._observedGridKeysByVortexUuid.items():
            for gridKey in gridKeys:
                newDict[gridKey].append(vortexUuid)

        keysChanged = set(self._observedVortexUuidsByGridKey) != set(newDict)

        self._observedVortexUuidsByGridKey = newDict

        # Notify the server that this client service is watching different grids.
        if keysChanged:
            d = ClientGridLoaderRpc.updateClientWatchedGrids(
                clientId=self._clientId,
                gridKeys=list(self._observedVortexUuidsByGridKey),
            )
            d.addErrback(vortexLogFailure, logger, consumeError=False)

    # ---------------
    # Reply to device observe

    def _replyToObserve(
        self,
        filt,
        lastUpdateByGridKey: DeviceGridT,
        sendResponse: SendVortexMsgResponseCallable,
        cacheAll=False,
    ) -> None:
        """Reply to Observe

        The client has told us that it's observing a new set of grids, and the lastUpdate
        it has for each of those grids. We will send them the grids that are out of date
        or missing.

        :param filt: The payload filter to respond to.
        :param lastUpdateByGridKey: The dict of gridKey:lastUpdate
        :param sendResponse: The callable provided by the Vortex (handy)
        :returns: None

        """
        gridTuplesToSend = []

        def sendChunk(toSend):
            if not toSend and not cacheAll:
                return

            payloadEnvelope = PayloadEnvelope(filt=filt, data=toSend)
            d: Deferred = payloadEnvelope.toVortexMsgDefer(base64Encode=False)
            d.addCallback(sendResponse)
            d.addErrback(vortexLogFailure, logger, consumeError=True)

        # Check and send any updates
        for gridKey, lastUpdate in lastUpdateByGridKey.items():
            # NOTE: lastUpdate can be null.
            gridTuple = self._cacheController.encodedChunk(gridKey)

            # Last update is not null, we need to send an empty grid.
            if not gridTuple:
                gridTuple = EncodedGridTuple()
                gridTuple.gridKey = gridKey
                gridTuple.lastUpdate = lastUpdate
                gridTuple.encodedGridTuple = None
                gridTuplesToSend.append(gridTuple)
                logger.debug(
                    "Grid %s is no loner in the cache, %s", gridKey, lastUpdate
                )

            elif gridTuple.lastUpdate == lastUpdate:
                logger.debug(
                    "Grid %s matches the cache, %s", gridKey, lastUpdate
                )

            else:
                gridTuplesToSend.append(gridTuple)
                logger.debug(
                    "Sending grid %s from the cache, %s", gridKey, lastUpdate
                )

            if len(gridTuplesToSend) == 5 and not cacheAll:
                sendChunk(gridTuplesToSend)
                gridTuplesToSend = []

        sendChunk(gridTuplesToSend)



from peek_plugin_base.storage.AlembicEnvBase import AlembicEnvBase

from peek_plugin_diagram._private.storage import DeclarativeBase
from peek_plugin_diagram._private.storage.DeclarativeBase import loadStorageTuples

loadStorageTuples()

alembicEnv = AlembicEnvBase(DeclarativeBase.DeclarativeBase.metadata)
alembicEnv.run()



from collections import namedtuple

import logging
from typing import Optional, List

from peek_plugin_diagram._private.tuples.location_index.DispKeyLocationTuple import (
    DispKeyLocationTuple,
)
from peek_plugin_diagram._private.worker.tasks._CalcGridForDisp import _calcBounds

logger = logging.getLogger(__name__)

CoordSetIdGridKeyTuple = namedtuple("CoordSetIdGridKeyTuple", ["coordSetId", "gridKey"])
DispData = namedtuple("DispData", ["json", "levelOrder", "layerOrder"])


def makeLocationJson(disp, geomArray: List[float]) -> Optional[str]:
    if geomArray is None:
        logger.critical("geomJson is None : %s ", disp)

    minx, miny, maxx, maxy = _calcBounds(geomArray)

    # Find the center point
    x = (minx + maxx) / 2
    y = (miny + maxy) / 2

    location = DispKeyLocationTuple(
        coordSetId=disp.coordSetId, dispId=disp.id, x=x, y=y
    )
    return location.toLocationJson()


def dispKeyHashBucket(modelSetKey: str, dispKey: str) -> str:
    """Disp Key Hash Bucket

    This method create an int from 0 to 255, representing the hash bucket for this
    key.

    This is simple, and provides a reasonable distribution

    :param modelSetKey:
    :param dispKey:

    :return:

    """
    if not modelSetKey:
        raise Exception("modelSetKey is None or zero length")

    if not dispKey:
        raise Exception("dispKey is None or zero length")

    hash = 0
    for char in dispKey:
        hash = ((hash << 5) - hash) + ord(char)
        hash = hash | 0  # This is in the javascript code.

    hash = hash & 1023  # 1024 buckets

    return "%s:%s" % (modelSetKey, hash)



"""Added overlay flag

Peek Plugin Database Migration Script

Revision ID: 3b41dec74e46
Revises: 97990b4ca6a1
Create Date: 2019-08-12 11:04:53.878245

"""

# revision identifiers, used by Alembic.
revision = "3b41dec74e46"
down_revision = "97990b4ca6a1"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.add_column(
        "DispBase",
        sa.Column("overlay", sa.Boolean(), server_default="false", nullable=True),
        schema="pl_diagram",
    )

    op.execute(
        """
        UPDATE pl_diagram."DispBase"
        SET overlay=sq.val
        FROM (
                SELECT id, "importGroupHash" is null or "importGroupHash" like '%.ov' as val
                FROM pl_diagram."DispBase"
        ) AS sq
        WHERE pl_diagram."DispBase".id=sq.id
    """
    )

    op.alter_column(
        "DispBase", "overlay", type_=sa.Boolean(), nullable=False, schema="pl_diagram"
    )

    op.execute(
        """
        INSERT INTO pl_diagram."DispCompilerQueue"("dispId")
        SELECT id FROM pl_diagram."DispBase"
    """
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("DispBase", "overlay", schema="pl_diagram")
    # ### end Alembic commands ###



"""Increased size of importGroupHash

Peek Plugin Database Migration Script

Revision ID: e9d80c4c087d
Revises: 4b943b584307
Create Date: 2017-10-25 10:03:26.694547

"""

# revision identifiers, used by Alembic.
revision = "e9d80c4c087d"
down_revision = "4b943b584307"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.alter_column(
        "DispBase",
        "importGroupHash",
        type_=sa.String(length=200),
        nullable=True,
        schema="pl_diagram",
    )


def downgrade():
    op.alter_column(
        "DispBase",
        "importGroupHash",
        type_=sa.String(length=100),
        nullable=True,
        schema="pl_diagram",
    )



"""add text border color and width

Peek Plugin Database Migration Script

Revision ID: e3c2c2580dea
Revises: 47879bec4c5f
Create Date: 2022-07-06 10:37:27.196092

"""

# revision identifiers, used by Alembic.
revision = "e3c2c2580dea"
down_revision = "47879bec4c5f"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    # add DispText.borderColorId and the foreign key
    op.add_column(
        "DispText",
        sa.Column("borderColorId", sa.Integer(), nullable=True),
        schema="pl_diagram",
    )
    op.create_foreign_key(
        "DispText_borderColorId_fkey",
        "DispText",
        "DispColor",
        ["borderColorId"],
        ["id"],
        source_schema="pl_diagram",
        referent_schema="pl_diagram",
    )

    # add DispTextStyle.borderWidth
    op.add_column(
        "DispTextStyle",
        sa.Column("borderWidth", sa.Float(), nullable=True),
        schema="pl_diagram",
    )


def downgrade():
    # drop DispText.borderColorId and the foreign key
    op.drop_constraint(
        "DispText_borderColorId_fkey", "DispText", schema="pl_diagram"
    )
    op.drop_column("DispText", "borderColorId", schema="pl_diagram")

    # drop DispTextStyle.borderWidth
    op.drop_column("DispTextStyle", "borderWidth", schema="pl_diagram")



from typing import List, Dict, Optional

from peek_plugin_diagram._private.storage.ModelSet import ModelSet
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler
from vortex.handler.TupleDataObserverClient import TupleDataObserverClient


class ModelSetCacheController:
    """Lookup Cache Controller

    This class caches the lookups in each client.

    """

    def __init__(self, tupleObserver: TupleDataObserverClient):
        self._tupleObserver = tupleObserver
        self._tupleObservable = None

        #: This stores the cache of grid data for the clients
        self._cache: Dict[int, ModelSet] = {}

    def setTupleObservable(self, tupleObservable: TupleDataObservableHandler):
        self._tupleObservable = tupleObservable

    def start(self):
        (
            self._tupleObserver.subscribeToTupleSelector(
                TupleSelector(ModelSet.tupleName(), {})
            ).subscribe(self._processNewTuples)
        )

    def shutdown(self):
        self._tupleObservable = None
        self._tupleObserver = None
        self._cache = {}

    def _processNewTuples(self, tuples):
        if not tuples:
            return

        self._cache = {c.id: c for c in tuples}

        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(ModelSet.tupleName(), {})
        )

    @property
    def modelSets(self) -> List[ModelSet]:
        return list(self._cache.values())



"""Encoded Grids are now GridTuples

Peek Plugin Database Migration Script

Revision ID: 298b291aac2d
Revises: c88ba8f6761f
Create Date: 2017-12-16 19:30:54.032544

"""

# revision identifiers, used by Alembic.
revision = "298b291aac2d"
down_revision = "c88ba8f6761f"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.execute(' TRUNCATE TABLE pl_diagram."GridKeyIndexCompiled" ')
    op.alter_column(
        "GridKeyIndexCompiled",
        "blobData",
        type_=sa.LargeBinary,
        new_column_name="encodedGridTuple",
        schema="pl_diagram",
    )
    op.execute(
        """INSERT INTO pl_diagram."GridKeyCompilerQueue" ("gridKey", "coordSetId")
                SELECT DISTINCT "gridKey", "coordSetId"
                FROM pl_diagram."GridKeyIndex"
                """
    )


def downgrade():
    raise Exception("Downgrade is not implemented")



"""Added landingCoordSetId

Peek Plugin Database Migration Script

Revision ID: 97990b4ca6a1
Revises: da424430bf97
Create Date: 2019-07-26 13:24:14.505584

"""

# revision identifiers, used by Alembic.
revision = "97990b4ca6a1"
down_revision = "da424430bf97"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "ModelSet",
        sa.Column("landingCoordSetId", sa.Integer(), nullable=True),
        schema="pl_diagram",
    )
    op.create_foreign_key(
        None,
        "ModelSet",
        "ModelCoordSet",
        ["landingCoordSetId"],
        ["id"],
        source_schema="pl_diagram",
        referent_schema="pl_diagram",
        ondelete="CASCADE",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "ModelSet", schema="pl_diagram", type_="foreignkey")
    op.drop_column("ModelSet", "landingCoordSetId", schema="pl_diagram")
    # ### end Alembic commands ###



"""Added polyline start/end types

Peek Plugin Database Migration Script

Revision ID: e8d6d02c0922
Revises: 6d3a96cf1955
Create Date: 2019-06-03 11:47:15.090106

"""

# revision identifiers, used by Alembic.
revision = "e8d6d02c0922"
down_revision = "6d3a96cf1955"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "DispPolyline",
        sa.Column("endEndType", sa.Integer(), nullable=True),
        schema="pl_diagram",
    )
    op.add_column(
        "DispPolyline",
        sa.Column("startEndType", sa.Integer(), nullable=True),
        schema="pl_diagram",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("DispPolyline", "startEndType", schema="pl_diagram")
    op.drop_column("DispPolyline", "endEndType", schema="pl_diagram")
    # ### end Alembic commands ###



from vortex.handler.TupleDataObservableProxyHandler import (
    TupleDataObservableProxyHandler,
)

from peek_plugin_diagram._private.client.controller.BranchIndexCacheController import (
    BranchIndexCacheController,
)
from peek_plugin_diagram._private.client.controller.CoordSetCacheController import (
    CoordSetCacheController,
)
from peek_plugin_diagram._private.client.controller.GridCacheController import (
    GridCacheController,
)
from peek_plugin_diagram._private.client.controller.LocationIndexCacheController import (
    LocationIndexCacheController,
)
from peek_plugin_diagram._private.client.controller.LookupCacheController import (
    LookupCacheController,
)
from peek_plugin_diagram._private.client.controller.ModelSetCacheController import (
    ModelSetCacheController,
)
from peek_plugin_diagram._private.client.tuple_providers.BranchIndexUpdateDateTupleProvider import (
    BranchIndexUpdateDateTupleProvider,
)
from peek_plugin_diagram._private.client.tuple_providers.BranchTupleProvider import (
    BranchTupleProvider,
)
from peek_plugin_diagram._private.client.tuple_providers.ClientCoordSetTupleProvider import (
    ClientCoordSetTupleProvider,
)
from peek_plugin_diagram._private.client.tuple_providers.ClientDispKeyLocationTupleProvider import (
    ClientDispKeyLocationTupleProvider,
)
from peek_plugin_diagram._private.client.tuple_providers.ClientGroupDispsTupleProvider import (
    ClientGroupDispsTupleProvider,
)
from peek_plugin_diagram._private.client.tuple_providers.ClientLocationIndexUpdateDateTupleProvider import (
    ClientLocationIndexUpdateDateTupleProvider,
)
from peek_plugin_diagram._private.client.tuple_providers.ClientLookupTupleProvider import (
    ClientLookupTupleProvider,
)
from peek_plugin_diagram._private.client.tuple_providers.ClientModelSetTupleProvider import (
    ClientModelSetTupleProvider,
)
from peek_plugin_diagram._private.client.tuple_providers.GridCacheIndexTupleProvider import (
    GridCacheIndexTupleProvider,
)
from peek_plugin_diagram._private.storage.Display import (
    DispLevel,
    DispTextStyle,
    DispLineStyle,
    DispColor,
    DispLayer,
)
from peek_plugin_diagram._private.storage.ModelSet import ModelCoordSet, ModelSet
from peek_plugin_diagram._private.tuples.GroupDispsTuple import GroupDispsTuple
from peek_plugin_diagram._private.tuples.branch.BranchIndexUpdateDateTuple import (
    BranchIndexUpdateDateTuple,
)
from peek_plugin_diagram._private.tuples.branch.BranchTuple import BranchTuple
from peek_plugin_diagram._private.tuples.grid.GridUpdateDateTuple import (
    GridUpdateDateTuple,
)
from peek_plugin_diagram._private.tuples.location_index.DispKeyLocationTuple import (
    DispKeyLocationTuple,
)
from peek_plugin_diagram._private.tuples.location_index.LocationIndexUpdateDateTuple import (
    LocationIndexUpdateDateTuple,
)


def makeClientTupleDataObservableHandler(
    tupleObservable: TupleDataObservableProxyHandler,
    modelSetCacheController: ModelSetCacheController,
    coordSetCacheController: CoordSetCacheController,
    gridCacheController: GridCacheController,
    lookupCacheController: LookupCacheController,
    locationCacheController: LocationIndexCacheController,
    branchCacheController: BranchIndexCacheController,
):
    """ " Make CLIENT Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param branchCacheController:
    :param modelSetCacheController:
    :param locationCacheController:
    :param tupleObservable: The tuple observable proxy
    :param coordSetCacheController: The clients coord set cache controller
    :param gridCacheController: The cache controller for the grid
    :param lookupCacheController: The clients lookup cache controller
    :return: An instance of :code:`TupleDataObservableHandler`

    """
    # Register TupleProviders here
    lookupTupleProvider = ClientLookupTupleProvider(lookupCacheController)

    # Add the lookup providers
    tupleObservable.addTupleProvider(DispLevel.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispLayer.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispColor.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispLineStyle.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispTextStyle.tupleName(), lookupTupleProvider)

    # Add the CoordSet providers

    tupleObservable.addTupleProvider(
        ModelSet.tupleName(), ClientModelSetTupleProvider(modelSetCacheController)
    )

    tupleObservable.addTupleProvider(
        ModelCoordSet.tupleName(), ClientCoordSetTupleProvider(coordSetCacheController)
    )

    tupleObservable.addTupleProvider(
        GridUpdateDateTuple.tupleName(),
        GridCacheIndexTupleProvider(gridCacheController),
    )

    tupleObservable.addTupleProvider(
        DispKeyLocationTuple.tupleName(),
        ClientDispKeyLocationTupleProvider(
            locationCacheController, coordSetCacheController
        ),
    )

    tupleObservable.addTupleProvider(
        LocationIndexUpdateDateTuple.tupleName(),
        ClientLocationIndexUpdateDateTupleProvider(locationCacheController),
    )

    tupleObservable.addTupleProvider(
        BranchTuple.tupleName(), BranchTupleProvider(branchCacheController)
    )

    tupleObservable.addTupleProvider(
        BranchIndexUpdateDateTuple.tupleName(),
        BranchIndexUpdateDateTupleProvider(branchCacheController),
    )

    tupleObservable.addTupleProvider(
        GroupDispsTuple.tupleName(), ClientGroupDispsTupleProvider(gridCacheController)
    )

    return tupleObservable



import logging

from peek_abstract_chunked_index.private.server.controller.ACIProcessorQueueControllerABC import (
    ACIProcessorQueueControllerABC,
    ACIProcessorQueueBlockItem,
)
from peek_abstract_chunked_index.private.server.controller.ACIProcessorStatusNotifierABC import (
    ACIProcessorStatusNotifierABC,
)
from peek_abstract_chunked_index.private.tuples.ACIProcessorQueueTupleABC import (
    ACIProcessorQueueTupleABC,
)
from peek_plugin_diagram._private.server.client_handlers.BranchIndexChunkUpdateHandler import (
    BranchIndexChunkUpdateHandler,
)
from peek_plugin_diagram._private.server.controller.StatusController import (
    StatusController,
)
from peek_plugin_diagram._private.storage.branch.BranchIndex import BranchIndex
from peek_plugin_diagram._private.storage.branch.BranchIndexCompilerQueue import (
    BranchIndexCompilerQueue,
)
from peek_plugin_diagram._private.storage.branch.BranchIndexEncodedChunk import (
    BranchIndexEncodedChunk,
)

logger = logging.getLogger(__name__)


class _Notifier(ACIProcessorStatusNotifierABC):
    def __init__(self, adminStatusController: StatusController):
        self._adminStatusController = adminStatusController

    def setProcessorStatus(self, state: bool, queueSize: int):
        self._adminStatusController.status.branchIndexCompilerQueueStatus = state
        self._adminStatusController.status.branchIndexCompilerQueueSize = queueSize
        self._adminStatusController.notify()

    def addToProcessorTotal(self, delta: int):
        self._adminStatusController.status.branchIndexCompilerProcessedTotal += delta
        self._adminStatusController.notify()

    def setProcessorError(self, error: str):
        self._adminStatusController.status.branchIndexCompilerLastError = error
        self._adminStatusController.notify()


class BranchIndexCompilerQueueController(ACIProcessorQueueControllerABC):
    QUEUE_ITEMS_PER_TASK = 10
    POLL_PERIOD_SECONDS = 1.000

    QUEUE_BLOCKS_MAX = 20
    QUEUE_BLOCKS_MIN = 4

    WORKER_TASK_TIMEOUT = 60.0

    _logger = logger
    _QueueDeclarative: ACIProcessorQueueTupleABC = BranchIndexCompilerQueue
    _VacuumDeclaratives = (
        BranchIndexCompilerQueue,
        BranchIndex,
        BranchIndexEncodedChunk,
    )

    def __init__(
        self,
        dbSessionCreator,
        statusController: StatusController,
        clientUpdateHandler: BranchIndexChunkUpdateHandler,
    ):
        ACIProcessorQueueControllerABC.__init__(
            self, dbSessionCreator, _Notifier(statusController)
        )

        self._clientUpdateHandler: BranchIndexChunkUpdateHandler = clientUpdateHandler

    def _sendToWorker(self, block: ACIProcessorQueueBlockItem):
        from peek_plugin_diagram._private.worker.tasks.branch.BranchIndexCompiler import (
            compileBranchIndexChunk,
        )

        return compileBranchIndexChunk.delay(block.itemsEncodedPayload)

    def _processWorkerResults(self, results):
        self._clientUpdateHandler.sendChunks(results)

    def _dedupeQueueSql(self, lastFetchedId: int, dedupeLimit: int):
        return """
                 with sq_raw as (
                    SELECT "id", "chunkKey"
                    FROM pl_diagram."BranchIndexCompilerQueue"
                    WHERE id > %(id)s
                    LIMIT %(limit)s
                ), sq as (
                    SELECT min(id) as "minId", "chunkKey"
                    FROM sq_raw
                    GROUP BY  "chunkKey"
                    HAVING count("chunkKey") > 1
                )
                DELETE
                FROM pl_diagram."BranchIndexCompilerQueue"
                     USING sq sq1
                WHERE pl_diagram."BranchIndexCompilerQueue"."id" != sq1."minId"
                    AND pl_diagram."BranchIndexCompilerQueue"."id" > %(id)s
                    AND pl_diagram."BranchIndexCompilerQueue"."chunkKey" = sq1."chunkKey"

            """ % {
            "id": lastFetchedId,
            "limit": dedupeLimit,
        }



"""added curved text disp

Peek Plugin Database Migration Script

Revision ID: a4c59bcaaf88
Revises: 8b8ebbe5ec7b
Create Date: 2022-05-27 14:30:03.965868

"""

# revision identifiers, used by Alembic.
revision = "a4c59bcaaf88"
down_revision = "8b8ebbe5ec7b"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        "DispCurvedText",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column(
            "verticalAlign", sa.Integer(), server_default="-1", nullable=False
        ),
        sa.Column(
            "horizontalAlign", sa.Integer(), server_default="0", nullable=False
        ),
        sa.Column(
            "text", sa.String(), server_default="new text label", nullable=True
        ),
        sa.Column("textHeight", sa.Float(), nullable=True),
        sa.Column("geomJson", sa.String(), nullable=False),
        sa.Column("colorId", sa.Integer(), nullable=True),
        sa.Column("textStyleId", sa.Integer(), nullable=False),
        sa.Column(
            "spacingBetweenTexts",
            sa.Float(),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["colorId"],
            ["pl_diagram.DispColor.id"],
        ),
        sa.ForeignKeyConstraint(
            ["id"], ["pl_diagram.DispBase.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["textStyleId"],
            ["pl_diagram.DispTextStyle.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="pl_diagram",
    )


def downgrade():
    op.drop_table("DispCurvedText", schema="pl_diagram")



from datetime import datetime
from typing import Dict

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from vortex.Tuple import addTupleType, TupleField, Tuple


@addTupleType
class GridUpdateDateTuple(Tuple):
    __tupleType__ = diagramTuplePrefix + "GridUpdateDateTuple"

    # Improve performance of the JSON serialisation
    __rawJonableFields__ = ("updateDateByChunkKey",)

    updateDateByChunkKey: Dict[str, str] = TupleField()



"""Removed BranchGridIndex

Peek Plugin Database Migration Script

Revision ID: bf8f559c4df9
Revises: 4f6ed1047562
Create Date: 2019-05-11 15:55:51.316969

"""

# revision identifiers, used by Alembic.
revision = "bf8f559c4df9"
down_revision = "4f6ed1047562"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "idx_BranchGridIndex_coord_key",
        table_name="BranchGridIndex",
        schema="pl_diagram",
    )
    op.drop_table("BranchGridIndex", schema="pl_diagram")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "BranchGridIndex",
        sa.Column(
            "id",
            sa.INTEGER(),
            server_default=sa.text(
                "nextval('pl_diagram.\"BranchGridIndex_id_seq\"'::regclass)"
            ),
            autoincrement=True,
            nullable=False,
        ),
        sa.Column("branchIndexId", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column("coordSetId", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column("gridKey", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.ForeignKeyConstraint(
            ["branchIndexId"],
            ["pl_diagram.BranchIndex.id"],
            name="BranchGridIndex_branchIndexId_fkey",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["coordSetId"],
            ["pl_diagram.ModelCoordSet.id"],
            name="BranchGridIndex_coordSetId_fkey",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", "coordSetId", name="BranchGridIndex_pkey"),
        schema="pl_diagram",
    )
    op.create_index(
        "idx_BranchGridIndex_coord_key",
        "BranchGridIndex",
        ["coordSetId", "branchIndexId", "gridKey"],
        unique=True,
        schema="pl_diagram",
    )
    # ### end Alembic commands ###



"""Added disp data

Peek Plugin Database Migration Script

Revision ID: d103b9c1eba9
Revises: 558bc7fc4d36
Create Date: 2017-09-23 14:58:37.938811

"""

# revision identifiers, used by Alembic.
revision = "d103b9c1eba9"
down_revision = "558bc7fc4d36"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.add_column(
        "DispBase",
        sa.Column("dataJson", sa.String(length=500), nullable=True),
        schema="pl_diagram",
    )


def downgrade():
    op.drop_column("DispBase", "dataJson", schema="pl_diagram")



from typing import List

from abc import ABCMeta, abstractmethod
from twisted.internet.defer import Deferred

from peek_plugin_diagram.server.DiagramImportApiABC import DiagramImportApiABC
from peek_plugin_diagram.server.DiagramViewerApiABC import DiagramViewerApiABC


class DiagramApiABC(metaclass=ABCMeta):
    def importApi(self) -> DiagramImportApiABC:
        """Import API

        :return: The Import API for the diagram plugin

        """

    def viewerApi(self) -> DiagramViewerApiABC:
        """Viewer API

        :return: The viewer API for the diagram plugin

        """



import logging
from typing import List

from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks

from peek_plugin_diagram._private.server.controller.BranchUpdateController import (
    BranchUpdateController,
)
from peek_plugin_diagram._private.server.controller.DispImportController import (
    DispImportController,
)
from peek_plugin_diagram._private.server.controller.LookupImportController import (
    LookupImportController,
)
from peek_plugin_diagram._private.server.controller.StatusController import (
    StatusController,
)
from peek_plugin_diagram._private.storage.Display import DispBase
from peek_plugin_diagram.server.DiagramImportApiABC import DiagramImportApiABC

logger = logging.getLogger(__name__)


class DiagramImportApi(DiagramImportApiABC):
    def __init__(
        self,
        mainController: StatusController,
        dispImportController: DispImportController,
        lookupImportController: LookupImportController,
        branchUpdateController: BranchUpdateController,
    ):
        self._mainController = mainController
        self._dispImportController = dispImportController
        self._lookupImportController = lookupImportController
        self._branchUpdateController = branchUpdateController

    def shutdown(self):
        pass

    def importDisps(
        self,
        modelSetKey: str,
        coordSetKey: str,
        importGroupHash: str,
        dispsEncodedPayload: bytes,
    ) -> Deferred:
        return self._dispImportController.importDisps(
            modelSetKey, coordSetKey, importGroupHash, dispsEncodedPayload
        )

    def importBranches(self, branchesEncodedPayload: bytes) -> Deferred:
        return self._branchUpdateController.importBranches(
            branchesEncodedPayload
        )

    def importLookups(
        self,
        modelSetKey: str,
        coordSetKey: str,
        lookupTupleType: str,
        lookupTuples: List,
        deleteOthers: bool = True,
        updateExisting: bool = True,
    ) -> Deferred:
        return self._lookupImportController.importLookups(
            modelSetKey,
            coordSetKey,
            lookupTupleType,
            lookupTuples,
            deleteOthers,
            updateExisting,
        )

    def getLookups(
        self, modelSetKey: str, coordSetKey: str, lookupTupleType: str
    ) -> Deferred:
        return self._lookupImportController.getLookups(
            modelSetKey, coordSetKey, lookupTupleType
        )

    def getImportGroupHashes(
        self, modelSetKey: str, coordSetKey: str, importGroupHashContains: str
    ) -> Deferred:
        return self._dispImportController.getImportGroupHashes(
            modelSetKey, coordSetKey, importGroupHashContains
        )

    def removeDispsByImportGroupHash(
        self, modelSetKey: str, coordSetKey: str, importGroupHash: str
    ) -> Deferred:
        yield self._dispImportController.removeDispsByImportGroupHash(
            modelSetKey, coordSetKey, importGroupHash
        )



import logging
from typing import Optional

from vortex.Tuple import Tuple

from peek_abstract_chunked_index.private.server.client_handlers.ACIChunkLoadRpcABC import (
    ACIChunkLoadRpcABC,
)
from peek_core_search._private.storage.EncodedSearchObjectChunk import (
    EncodedSearchObjectChunk,
)
from peek_plugin_base.PeekVortexUtil import peekServerName, peekBackendNames
from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.storage.LocationIndex import (
    LocationIndexCompiled,
)
from peek_plugin_diagram._private.storage.ModelSet import ModelSet
from sqlalchemy import select
from vortex.rpc.RPC import vortexRPC

logger = logging.getLogger(__name__)


class ClientLocationIndexLoaderRpc(ACIChunkLoadRpcABC):
    def makeHandlers(self):
        """Make Handlers

        In this method we start all the RPC handlers
        start() returns an instance of it's self so we can simply yield the result
        of the start method.

        """

        yield self.loadLocationIndexes.start(funcSelf=self)
        logger.debug("RPCs started")

    # -------------
    @vortexRPC(
        peekServerName,
        acceptOnlyFromVortex=peekBackendNames,
        timeoutSeconds=60,
        additionalFilt=diagramFilt,
        deferToThread=True,
    )
    def loadLocationIndexes(self, offset: int, count: int) -> str:
        """Update Page Loader Status

        Tell the server of the latest status of the loader

        """
        chunkTable = LocationIndexCompiled.__table__
        msTable = ModelSet.__table__

        sql = (
            select(
                [
                    chunkTable.c.indexBucket,
                    chunkTable.c.blobData,
                    chunkTable.c.lastUpdate,
                    msTable.c.key,
                ]
            )
            .select_from(chunkTable.join(msTable))
            .order_by(chunkTable.c.indexBucket)
            .offset(offset)
            .limit(count)
        )

        return self.ckiInitialLoadChunksPayloadBlocking(
            offset, count, LocationIndexCompiled, sql
        )



"""coordSet.positionOnZoom

Peek Plugin Database Migration Script

Revision ID: a73574690197
Revises: 3b41dec74e46
Create Date: 2019-08-12 13:06:03.511202

"""

# revision identifiers, used by Alembic.
revision = "a73574690197"
down_revision = "3b41dec74e46"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.add_column(
        "ModelCoordSet",
        sa.Column("positionOnZoom", sa.Float(), server_default="2.0", nullable=True),
        schema="pl_diagram",
    )

    op.execute(
        """
        UPDATE pl_diagram."ModelCoordSet"
        SET "positionOnZoom"=2.0
    """
    )

    op.alter_column(
        "ModelCoordSet",
        "positionOnZoom",
        type_=sa.Float(),
        nullable=False,
        schema="pl_diagram",
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("ModelCoordSet", "positionOnZoom", schema="pl_diagram")
    # ### end Alembic commands ###



import logging

from peek_abstract_chunked_index.private.server.controller.ACIProcessorStatusControllerABC import (
    ACIProcessorStatusControllerABC,
)
from peek_plugin_diagram._private.tuples.DiagramImporterStatusTuple import (
    DiagramImporterStatusTuple,
)

logger = logging.getLogger(__name__)


class StatusController(ACIProcessorStatusControllerABC):
    NOTIFY_PERIOD = 2.0

    _StateTuple = DiagramImporterStatusTuple



from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from vortex.Tuple import addTupleType, TupleField, Tuple


@addTupleType
class DiagramImporterStatusTuple(Tuple):
    __tupleType__ = diagramTuplePrefix + "DiagramImporterStatusTuple"

    displayCompilerQueueStatus: bool = TupleField(False)
    displayCompilerQueueSize: int = TupleField(0)
    displayCompilerProcessedTotal: int = TupleField(0)
    displayCompilerLastError: str = TupleField()

    gridCompilerQueueStatus: bool = TupleField(False)
    gridCompilerQueueSize: int = TupleField(0)
    gridCompilerProcessedTotal: int = TupleField(0)
    gridCompilerLastError: str = TupleField()

    locationIndexCompilerQueueStatus: bool = TupleField(False)
    locationIndexCompilerQueueSize: int = TupleField(0)
    locationIndexCompilerProcessedTotal: int = TupleField(0)
    locationIndexCompilerLastError: str = TupleField()

    branchIndexCompilerQueueStatus: bool = TupleField(False)
    branchIndexCompilerQueueSize: int = TupleField(0)
    branchIndexCompilerProcessedTotal: int = TupleField(0)
    branchIndexCompilerLastError: str = TupleField()



"""Removed Node and Conn

Peek Plugin Database Migration Script

Revision ID: 2c5f47322bda
Revises: d6e25d8ba156
Create Date: 2017-08-24 12:58:59.016532

"""

# revision identifiers, used by Alembic.
revision = "2c5f47322bda"
down_revision = "d6e25d8ba156"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("DispPolylineConn", schema="pl_diagram")
    op.drop_table("DispGroupPointerNode", schema="pl_diagram")
    op.drop_table("ModelConn", schema="pl_diagram")
    op.drop_table("ModelNode", schema="pl_diagram")
    op.drop_table("DispAction", schema="pl_diagram")
    op.drop_table("ModelNodeType", schema="pl_diagram")
    op.drop_table("ModelConnType", schema="pl_diagram")

    op.add_column(
        "DispBase",
        sa.Column("key", sa.String(length=50), nullable=True),
        schema="pl_diagram",
    )
    op.add_column(
        "DispBase",
        sa.Column("selectable", sa.Boolean(), server_default="0", nullable=False),
        schema="pl_diagram",
    )

    op.add_column(
        "ModelSet",
        sa.Column("key", sa.String(length=50), nullable=True),
        schema="pl_diagram",
    )
    op.execute(""" UPDATE "pl_diagram"."ModelSet" SET "key" = "name" """)
    op.alter_column(
        "ModelSet",
        "key",
        type_=sa.String(length=50),
        nullable=False,
        schema="pl_diagram",
    )

    # ### end Alembic commands ###


def downgrade():
    raise NotImplementedError("Downgrade is not implemented")



"""moved smallest sizes to grids

Peek Plugin Database Migration Script

Revision ID: 2127d894e46b
Revises: 6a97dd344aed
Create Date: 2017-08-13 21:31:51.165359

"""

# revision identifiers, used by Alembic.
revision = "2127d894e46b"
down_revision = "6a97dd344aed"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2
from sqlalchemy.dialects import postgresql


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("ModelCoordSet", "smallestTextSize", schema="pl_diagram")

    op.drop_column("ModelCoordSet", "smallestShapeSize", schema="pl_diagram")

    op.add_column(
        "ModelCoordSetGridSize",
        sa.Column("smallestTextSize", sa.Float(), nullable=True, server_default="6.0"),
        schema="pl_diagram",
    )

    op.add_column(
        "ModelCoordSetGridSize",
        sa.Column("smallestShapeSize", sa.Float(), nullable=True, server_default="2.0"),
        schema="pl_diagram",
    )

    op.execute(
        """
    UPDATE "pl_diagram"."ModelCoordSetGridSize"
    SET "smallestShapeSize" = 2.0,
        "smallestTextSize" = 6.0
    """
    )

    op.alter_column(
        "ModelCoordSetGridSize",
        "smallestTextSize",
        existing_type=sa.Float(),
        nullable=False,
        schema="pl_diagram",
    )

    op.alter_column(
        "ModelCoordSetGridSize",
        "smallestShapeSize",
        existing_type=sa.Float(),
        nullable=False,
        schema="pl_diagram",
    )


def downgrade():
    raise NotImplementedError("Downgrade is not implemented")
